# -*- coding: utf-8 -*-
# Copyright 2010 Intevation GmbH, Germany, <info@intevation.de>
# Authors:
# Bernhard Herzog <bernhard.herzog@intevation.de>
#

import sys
import re
from datetime import date
from xml.sax.handler import ContentHandler
from xml.sax import make_parser

import formed.meta.data as data
from formed.meta.data import InfoLeaf
import formed.meta.semantic as semantic
from formed.meta.structure import StructureTree
from formed.instance.simplified import StructureInstanceNode, \
     StructureInstanceTree
from formed.instance.tree import RepeatInstanceContainerNode, SetDataException
from formed.instance.backends.postgres import DBFactory


class XMLImportException(Exception):

    """Raised by the import code if there's a problem.
    Instances have the following public attributes:

      message -- A string with a description of the exception

      uuid -- A string with the uuid of the case whose import failed

      set_data_errors -- A dictionary with the error descriptions from a
                         SetDataException
    """

    def __init__(self, message, uuid=None, set_data_errors=()):
        Exception.__init__(self, message)
        self.uuid = uuid
        self.set_data_errors = set_data_errors


re_date = re.compile("([0-9]{1,4})-([0-9]{1,2})-([0-9]{1,2})")


def decode_int(s):
    return int(s)

def decode_date(s):
    m = re_date.match(s)
    if not m:
        raise ValueError("%r cannot be decoded as a date" % s)
    return date(int(m.group(1)), int(m.group(2)), int(m.group(3)))

def decode_text(s):
    return s


DECODERS = {
    data.BoolLeaf: decode_int,
    data.ChoiceNode: decode_int,
    data.DateLeaf: decode_date,
    data.IntLeaf: decode_int,
    data.RadioNode: decode_int,
    data.TextAreaLeaf: decode_text,
    data.TextLeaf: decode_text,
    }


def decode(meta_node, value):
    decoder = DECODERS.get(meta_node.__class__)
    if decoder is not None:
        value = decoder(value)
    else:
        print >>sys.stderr, "decode: no decoder for %r:%r" % (meta_node, value)
    return semantic.check(meta_node, value)


class CasesHandler(ContentHandler):

    def __init__(self, xml_importer):
        self.xml_importer = xml_importer
        self.cases = None
        self.case = None

    def startElement(self, name, attrs):
        if name == "cases":
            if self.cases is not None:
                raise XMLImportException("unexpected 'cases' element")
            if attrs.get("version") != "5":
                raise XMLImportException("Only version 5 of the cases"
                                         " element is supported")
            self.xml_importer.version_supported(attrs.get("formed-type"))
            self.cases = []
        elif name == "case":
            if self.case is not None:
                raise XMLImportException("unexpected 'case' element")
            self.case = CaseHandler(self.xml_importer.case_factories)
        else:
            if self.case is not None:
                self.case.startElement(name, attrs)
            else:
                raise XMLImportException("unexpected element: %r" % name)

    def endElement(self, name):
        if name == "cases":
            pass
        elif name == "case":
            if self.case is None:
                raise XMLImportException("unexpected 'case' end-tag")
            else:
                self.cases.append(self.case.get_instance())
                self.case = None
        else:
            if self.case is not None:
                self.case.endElement(name)
            else:
                raise XMLImportException("unexpected end-tag: %r" % name)

    def characters(self, content):
        if self.case:
            self.case.characters(content)


class ComponentFactory(object):

    """Factories to create StructureTree based trees."""

    def __init__(self, meta_tree, meta_root_node, node_factory, tree_factory):
        self.meta_tree = meta_tree
        self.meta_root_node = meta_root_node
        self.node_factory = node_factory
        self.tree_factory = tree_factory


class CaseHandler(ContentHandler):

    def __init__(self, component_factories):
        """Initialize the CaseHandler.

        The component_factories parameter should be a dictionary mapping
        names of child elements of the case element to the corresponding
        ComponentFactory instance."""
        self.component_factories = component_factories
        self.components = dict()
        self.nodestack = []
        self.cur_meta = self.cur_leaves = self.cur_rgs = None
        self.text_content = None
        self.factory = None

    def get_instance(self):
        """Return the instance read from the XML file.
        The return value is a dictionary mapping the keys from the
	component_factories dictionary given to the constructor to the
	instance tree created for that component if it actually was
	present in the XML file.
        """
        return self.components

    def push_node(self, node):
        self.cur_meta = node
        self.cur_leaves = dict()
        self.cur_rgs = dict()
        self.nodestack.append((self.cur_meta, self.cur_leaves, self.cur_rgs))

    def pop_node(self):
        node, leaves, rgs = self.nodestack.pop()
        for leafname in node.leaves:
            if leafname not in leaves:
                leaves[leafname] = None
        for rgname in node.repeat_groups:
            if rgname not in rgs:
                rgs[rgname] = dict()
        if self.nodestack:
            self.cur_meta, self.cur_leaves, self.cur_rgs = self.nodestack[-1]
        else:
            self.cur_meta = self.cur_leaves = self.cur_rgs = None
        return self.create_instance_node(node, leaves, rgs)

    def startElement(self, name, attrs):
        if name in self.component_factories:
            self.factory = self.component_factories[name]
            if self.nodestack:
                raise XMLImportException("Unexpected element %r" % name)
            self.push_node(self.factory.meta_root_node)
        else:
            if name in self.cur_meta.leaves or name == "uuid_id":
                self.text_content = []
            elif name in self.cur_meta.repeat_groups:
                if self.text_content is not None:
                    raise XMLImportException("Unexpected element: %r" % name)
                self.push_node(self.cur_meta.repeat_groups[name])
            else:
                raise XMLImportException("Unexpected element: %r" % name)

    def endElement(self, name):
        if name in self.component_factories:
            assert(self.component_factories[name] is self.factory)
            if len(self.nodestack) != 1:
                raise XMLImportException("Unexpected element %r" % name)
            node = self.pop_node()
            if not node.meta_node.is_root():
                raise XMLImportException("node %r is not root node" % name)
            self.components[name] = \
                     self.factory.tree_factory(self.factory.meta_tree, node)
        else:
            if name in self.cur_meta.leaves:
                self.cur_leaves[name] = decode(self.cur_meta.leaves[name],
                                               self.get_text_content())
            elif name == "uuid_id":
                self.cur_leaves[name] = self.get_text_content()
            elif name == self.cur_meta.name:
                if self.text_content is not None:
                    raise XMLImportException("Unexpected element: %r" % name)
                new_node = self.pop_node()
                # we don't have database ids for the repeat group
                # instances, so invent one that is unique enough: the
                # number of instances already created
                instances = self.cur_rgs.setdefault(name, {})
                instances[len(instances)] = new_node
            else:
                raise XMLImportException("Unexpected element: %r" % name)

    def create_instance_node(self, structure_node, leaves, repeat_groups):
        # uuids are optional and only present if cases have not yet been
        # anonymized, either by anonymizing the case in the database or
        # by exporting an anonymized version.
        uuid = leaves.pop("uuid_id", None)
        return self.factory.node_factory(structure_node, None, uuid, leaves,
                                         repeat_groups)

    def get_text_content(self):
        content = "".join(self.text_content)
        self.text_content = None
        return content

    def characters(self, content):
        if self.text_content is not None:
            self.text_content.append(content)


def update_instance_in_database(src, factory):
    """Update or create the database instance corresponding to src.

    The parameter src must be a structure instance tree.  If the case
    described by src already exists in the database as determined by
    their uuids, that case is updated.  Otherwise a new case is created
    in the database.

    The return value is the instance tree the case was stored into.
    """
    uuid = src.uuid
    dest_tree = None
    match_uuids = False
    if uuid:
        dest_tree = factory.loadInstanceTreeByUUID(uuid)
        match_uuids = True
    if dest_tree is None:
        dest_tree = factory.createNewInstanceTree(uuid)
    copy_to_full_instance_tree(src, dest_tree, match_uuids=match_uuids)
    return dest_tree


def copy_to_full_instance_tree(src, dest, match_uuids):
    data = dict()
    _copy_data_recursively(src.root, dest.getRootNode(), data,
                           match_uuids=match_uuids)
    try:
        dest.setData(data, convert=False, find_best_changes=False)
    except SetDataException, e:
        print >>sys.stderr, "SetDataException when loading a case from XML", e
        raise XMLImportException("The case with uuid '%s' could not be imported"
                                 " due to rule violations " % src.uuid,
                                 uuid=src.uuid, set_data_errors=e.errors)
    dest.commit()
    return dest

def _copy_data_recursively(src, dest, data, match_uuids):
    dest_children = dest.getChildren()

    if isinstance(dest, RepeatInstanceContainerNode):
        src_rgs = src.get_repeat_groups(dest.getMeta().getName())
        dest_mapping = dict((child.getUUID(), child) for child in dest_children)
        for child_id, src_child in src_rgs.iteritems():
            # FIXME: use uuids here
            dest_child = dest_mapping.pop(src_child.uuid, None)
            if dest_child is None or not match_uuids:
                dest_child = dest.createInstance(src_child.uuid)
            _copy_data_recursively(src_child, dest_child, data,
                                   match_uuids=match_uuids)
        for child_id, dest_child in dest_mapping.iteritems():
            dest.deleteInstance(dest_child)
    else:
        for child in dest_children:
            if child.isLeaf():
                if isinstance(child.getMeta(), InfoLeaf):
                    continue
                value = src.get_leaf(child.getMeta().getName())
                if value is not None:
                    data[child.getIdentifier()] = value
            else:
                _copy_data_recursively(src, child, data,
                                       match_uuids=match_uuids)



class XMLImporter(object):

    def __init__(self, meta_tree, extra_factories=None):
        """Initialize the importer.
        The meta_tree parameter should be the formed meta tree matching
        the data to be read.
        """
        self.structure_tree = StructureTree(meta_tree)
        factories = {"master": ComponentFactory(self.structure_tree,
                                                self.structure_tree.root,
                                                StructureInstanceNode,
                                                StructureInstanceTree)}
        if extra_factories is not None:
            factories.update(extra_factories)
        self.case_factories = factories

    def import_xml(self, db, infile):
        """Read cases from the file object infile and update the database db.
        The db parameter must conform to the interface
        formed.instance.backends.database.DB.

        The import_case method is called once for each instance tree
        read from the xml file.  It is called with the instance tree
        factory, a database cursor and the structure instance tree.
        Derived classes may want to override that method.

        All database changes are done in a transaction.  If any exception is
        raised during the import, that transaction is rolled back, so no
        database changes will be made.
        """
        factory = DBFactory(self.structure_tree.meta_tree, db)
        db.begin()
        cursor = None
        try:
            try:
                cursor = db.acquire_cursor()
                for case in self.read_cases_from_xml(infile):
                    self.import_case(factory, cursor, case)
                db.commit()
            except:
                db.rollback()
                raise
        finally:
            db.release_cursor(cursor)

    def read_cases_from_xml(self, infile, chunksize=16*1024):
        """Read cases from the infile and yield each once it has been read.
        """
        handler = CasesHandler(self)
        parser = make_parser()
        parser.setContentHandler(handler)

        while 1:
            chunk = infile.read(chunksize)
            if not chunk:
                break
            parser.feed(chunk)
            while handler.cases:
                yield handler.cases.pop(0)

    def import_case(self, factory, cursor, components):
        """Write the case into the database using the given db cursor.
        The factory parameter is a DBFactory object that can be used to
        create new cases in the database.  The cursor is a cursor for
        the same database connection used by the factory so that all
        changes made through the cursor are part of the same transaction
        as the changes done via the factory.  The commit -- or rollback
        in the case of errors -- will be done by the caller of this
        method.

        The return value is the new or updated instance tree object.

        Derived classes may override or extend this method to
        e.g. determine whether to import a particular case at all.
        """
        return update_instance_in_database(components["master"], factory)

    def version_supported(self, version):
        """Return whether the given version of the formed tree is supported.
        Derived class should override this to check the version and
        raise an XMLImportException if the version is not acceptable.
        """
        return True
