#!/usr/bin/env python
#
# (c) 2009 by Sascha L. Teichmann
#

import sys
import os.path
import getopt
import time
import socket
import traceback

from signal import SIGTERM

from datetime import datetime

from threading import Thread, Condition

from libmpuls.service.tags import TagSetterConfig, TagSetter

DIR = "/"

LOG_FILE = "/var/log/tagd.log"
PID_FILE = "/var/run/tagd.pid"

START_MSG = "Started tagd with pid"

STDOUT = "/dev/null"
STDERR = None
STDIN  = "/dev/null"

UMASK = 0007

UID = None
GID = None

UNIX_SOCKET = "/tmp/tagd-socket"

USAGE = \
'''usage: %s [<options>] {start|stop|restart}
where options are:
-h|--help          This text
-o|--out=<file>    redirect stdout to file  (def: /dev/null)
-e|--err=<file>    redirect stderr to file  (def: /dev/null)
-i|--in=<file>     redirect stdin  to file  (def: /dev/null)
-l|--log=<file>    append log to file       (def: /var/log/tagd.log)
-p|--pid=<file>    write pid to file        (def: /var/run/tagd.pid)
-d|--dir<dir>      change working directory (def: /)
-u|--uid=<uid>     set effective user id    (def: no change)
-g|--gid=<gid>     set effective group id   (def: no change)
-m|--umask=<umask> set umask                (def: 0002)
-c|--socket=<file> UNIX socket to listen on (def: /tmp/tagd-socket)
-s|--start<msg>    set start message
                   (def: 'Started tagd with pid')
'''

class Parameters(object):

    def __init__(self):
        self.stdout      = STDOUT
        self.stderr      = STDERR
        self.stdin       = STDIN
        self.umask       = UMASK
        self.uid         = UID
        self.gid         = GID
        self.pid_file    = PID_FILE
        self.log_file    = LOG_FILE
        self.dir         = DIR
        self.start_msg   = START_MSG
        self.unix_socket = UNIX_SOCKET

class Worker(Thread):
    
    def __init__(self, params, service):
        Thread.__init__(self)
        self.params  = params
        self.service = service

    def run(self):


        tag_config = '/var/lib/postgresql/tagging.xml'
        username = 'postgres'
        password = ''
        db = None
        host = None
        port = 5432
        caseid = None

        # Initial TagsetterConfig. Parts will be overwritten later 
        config = TagSetterConfig(tag_config, username, password, db, host, port, caseid)
        current_db = None

        while True:
            self.service.cv.acquire()
            while not self.service.queue:
                self.service.cv.wait()
            item = self.service.queue.pop(0)
            self.service.cv.release()

            db, caseid = item.split(' ')
            if db != current_db:
                # set new db connection in config and Create new Tagsetter
                config.set_db(username, password, db, host, port) 
                setter = TagSetter(config)
                current_db = db
            if current_db:
                setter.perform(caseid)


class Service(object):

    def __init__(self):
        self.socket = None
        self.cv     = Condition()
        self.queue  = []

    def enqueue(self, item):
        self.cv.acquire()
        self.queue.append(item)
        self.cv.notify()
        self.cv.release()

    def service(self, params):
        '''this is daemonized part'''
        
        try:
            os.remove(params.unix_socket)
        except OSError:
            pass

        worker = Worker(params, self)
        worker.setDaemon(True)
        worker.start()

        self.socket = None
        try:
            self.socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            self.socket.bind(params.unix_socket)

            self.socket.listen(1)

            while True:
                conn, addr = self.socket.accept()

                in_buf = []
                while len(in_buf) < 100:
                    data = conn.recv(1024)
                    if not data: break
                    in_buf.append(data)

                in_buf = ''.join(in_buf).strip()
                if in_buf: self.enqueue(in_buf)
                try:
                    con.close()
                except:
                    pass

        finally:
            if self.socket:
                try: self.socket.close()
                except: pass
                self.socket = None

            try:
                os.remove(params.unix_socket)
            except OSError:
                pass


def die(msg, exit_code = 1):
    log(msg)
    sys.exit(exit_code)

def log(msg):
    import syslog
    syslog.openlog("%s[%d]" % (
        os.path.basename(sys.argv[0]), os.getpid()),
        0,
        syslog.LOG_DAEMON)
    syslog.syslog(syslog.LOG_ERR, msg)
    syslog.closelog()

def daemonize(service, params):

    try:
        if os.fork() > 0: sys.exit(0)
    except OSError, e:
        die("fork #1 failed: (%d) %s" % (e.errno, e.strerror))

        os.chdir(params.dir)
        if params.umask is not None: os.umask(params.umask)
        os.setsid()

    try:
        pid = os.fork()
    except OSError, e:
        die("fork #2 failed: (%d) %s\n" % (e.errno, e.strerror))

    if pid > 0: 
        print >> sys.stderr, params.start_msg, pid
        if params.pid_file:
            pfile = None
            try:
                pfile = open(params.pid_file, "w")
                print >> pfile, "%d" % pid
            finally:
                if pfile: pfile.close()
        sys.exit(0)

    if not params.stderr: params.stderr = params.stdout

    sys.stderr.flush()
    sys.stdout.flush()

    try:
        if params.gid is not None:
            os.setgid(params.gid)

        if params.uid is not None:
            os.setuid(params.uid)

        si = open(params.stdin,  "r")
        so = open(params.stdout, "a+")
        se = open(params.stderr, "a+", 0)

        os.dup2(si.fileno(), sys.stdin.fileno())
        os.dup2(so.fileno(), sys.stdout.fileno())
        os.dup2(se.fileno(), sys.stderr.fileno())

        service(params)
    except:
        die(traceback.format_exc())

def usage(msg = "", exit_code = 0):
    if msg: print >> sys.stderr, msg
    print >> sys.stderr, USAGE % sys.argv[0]
    sys.exit(exit_code)

def main():

    try:
        opts, args = getopt.getopt(sys.argv[1:], 
            "ho:e:i:l:p:d:u:g:m:s:c:", 
            ["help", "out=", "err=", "in=", "log=", "pid=",
             "dir=", "uid=", "gid=", "umask=", "start=",
             "socket=" ])
    except getopt.GetoptError, err:
        usage(str(err), 2)

    params = Parameters()

    for o, a in opts:
        if o in ('h', '-h', '--help'):
            usage(0)
        elif o in ('o' '-o', '--out'):
            params.stdout = a
        elif o in ('i', '-i', '--in'):
            params.stdin = a
        elif o in ('l', '-l', '--log'):
            params.log_file = a
        elif o in ('p', '-p', '--pid'):
            params.pid_file = a
        elif o in ('d', '-d', '--dir'):
            params.dir = a
        elif o in ('s', '-s', '--start'):
            params.start_msg = a
        elif o in ('c', '-c', '--socket'):
            params.unix_socket = a
        elif o in ('u', '-u', '--uid'):
            try:
                params.uid = int(a)
                if (params.uid < 0): raise ValueError()
            except ValueError:
                usage("error: uid is not a positive integer.", 1)
        elif o in ('g', '-g', '--gid'):
            try:
                params.gid = int(a)
                if (params.gid < 0): raise ValueError()
            except ValueError:
                usage("error: gid is not a positive integer.", 1)
        elif o in ('m', '-m', '--umask'):
            try:
                params.umask = int(a, 8)
                if (params.umask < 0): raise ValueError()
            except ValueError:
                usage("error: umask is not valid octal integer.", 1)

    if not args:
        usage("error: missing action", 1)

    actions = None

    if   args[0] == "start":   actions = [ "start" ]
    elif args[0] == "stop":    actions = [ "stop"  ]
    elif args[0] == "restart": actions = [ "stop", "start"]

    if not actions:
        usage("error: No valid action", 1)

    for action in actions:
        if action == "stop":
            if not os.path.isfile(params.pid_file):
                print >> sys.stderr, "warning: pid file '%s' not found" \
                    % params.pid_file
            else:
                pid, pfile = None, None
                try:
                    try:
                        pfile = open(params.pid_file, "r")
                        pid = int(pfile.read().strip())
                        if pid < 0: raise ValueError()
                    finally:
                        if pfile: pfile.close()
                except (IOError, ValueError):
                    print >> sys.stderr, "warning: pid is not readable"

                if pid is not None:
                    while xrange(300):
                        try:
                            os.kill(pid, SIGTERM)
                        except OSError:
                            break
                        time.sleep(1)
                os.remove(params.pid_file)
        elif action == "start":
            if os.path.isfile(params.pid_file):
                usage("error: pid file '%s' exists." % params.pid_file, 1)
            service = Service()
            daemonize(service.service, params)


if __name__ == '__main__':
    main()

# vim: et si ai sw=4 ts=4 bs=2 fenc=utf-8 :
