# -*- coding: utf-8 -*-
from datetime import datetime
import logging
import formencode

from pylons import request, tmpl_context as c, app_globals as g
from pylons.controllers.util import redirect_to

from formed.instance.tree import SetDataException

from mpulsweb.lib.translation import _, ungettext
from mpulsweb.lib.base import BaseController, render, session, h
from mpulsweb.lib.security import checkRole
from mpulsweb.lib.dialogs import checkPrivacyStatement
from mpulsweb.lib.validators import SetStandinForm, SetEditorForm, NewCaseForm
from mpulsweb.lib.anonymize import determineAnonymizedCaseEndPhaseName, \
determineAnonymizedCaseEndPhase
from mpulsweb.lib.db import  PostgresDBInterface
from mpulsweb.model.case import ConsistenceCheckException, SessionCase, \
     MPulsXMLImporter, XMLImportException
from mpulsweb.model.user import UserGroupList, UserListObject
from mpulsweb.model.logbook import LogbookEntry
from mpulsweb.controllers.formularpage import convert2dic

from mpulsweb.controllers.formularpage import get_rendered_page

ANONYMIZE_CONFIRM = u"""Fallakte anonymisieren?"""
MARKANONYMIZE_CONFIRM_TEXT = u"""\
<p>Wollen Sie wirklich die Fallakte zum Anonymisieren freigeben? Die
Fallakte steht Ihnen danach nicht mehr zur Bearbeitung zur Verfügung und
wird der Administration zur Anonymisierung vorgelegt.</p>
<p><b>Die Fallakte wird in der Phase </b>'%s'<b> anonymisiert!</b>"""
ANONYMIZE_CONFIRM_TEXT = _("""Do you really want to anonymise the case and
irrecovably lose the associated personal data? After it the case will
not be available you. The anonymised case will still be considered for
the analysis.""")
ANONYMIZE_SUCCESS = u"""Fallakte anonymisiert!"""
ANONYMIZE_SUCCESS_TEXT = u"""\
Die personenbezogenen Daten der Fallakte wurden gelöscht und die
Fallakte aus der Übersicht entfernt. Bitte klicken Sie auf "OK",
um zur Fallaktenübersicht fortzufahren."""
MARKANONYMIZE_FAILED = u"""Anonymisierung nicht möglich."""
MARKANONYMIZE_FAILED_TEXT = u""" <p>Eine Anonymisierung der Fallakte war aus
folgendem Grund nicht möglich:</p><p><i>%s</i></p><p>Bitte klicken Sie auf "OK",
um zur Fallaktenübersicht fortzufahren.</p>"""
MARKANONYMIZE_SUCCESS = u"""Fallakte wurde zur Anonymisierung freigegeben!"""
MARKANONYMIZE_SUCCESS_TEXT = u"""\
Die Fallakte wurde zur Anonymisierung freigeben und liegt der
Administration zur endgültigen Anonymisierung vor. Bitte klicken Sie auf "OK",
um zur Fallaktenübersicht fortzufahren."""

# Use ungettext here because in the po file this translation is ready for
# ungettext. As the case controller only handles single cases provide constant
# 1
DELETE_CONFIRM = ungettext('Delete case?', 'Delete cases?', 1)
DELETE_CONFIRM_TEXT = _("""Warning! On deleting the case all data will be erased
irrecoverable!<br>Only delete the case if this case is <b>not needed for
evaluations</b> anymore.<br><br>Do you want to delete the case from the
server?""")
MARKDELETE_CONFIRM_TEXT = _("""Do you really want to mark this case for deletion?
The case will not be editable afterwards, and submited to the administration
for deletion.""")

# Use ungettext here because in the po file this translation is ready for
# ungettext. As the case controller only handles single cases provide constant
# 1
RESTORE_CONFIRM = ungettext('Restore case?', 'Restore cases?', 1)
RESTORE_CONFIRM_TEXT = _("""Do you really want to restore the case? The case
will be editable afterwards again for the users.""")

IMPORT_XML_SUCCESS = _("Case(s) imported!""")
IMPORT_XML_SUCCESS_TEXT = _("""The file was imported successfully.
Click on "OK" to open the case overview.""")

IMPORT_FAILED = _("Case(s) not imported!")
IMPORT_FAILED_TEXT = _("""An error occurred during the import.
Click 'OK' to return to the import dialog.""")


log = logging.getLogger(__name__)

class CaseController(BaseController):

    def init_case(self, id):
        '''Return a case object'''
        case = self._loadCase(id)
        return case

    @checkRole(('admin', 'cm', 'pt_dlr'))
    def index(self):
        redirect_to(controller="/case_overview")

    @checkRole(('admin', 'cm', 'pt_dlr'))
    @checkPrivacyStatement()
    def select(self, id, confirmed=0):
        return self.digest(id)

    @checkRole(('admin', 'cm', 'pt_dlr'))
    def digest(self, id):
        id = self._checkInt(id)
        case = self.init_case(id)
        c.case = case
        c.last_access = case.getState().getAccessTime()
        c.daysToMaxSavetime = case.getState().daysToMaxSavetime()
        c.maxsavetime = case.getState().getMaxSavetime()
        c.finished = case.getState().isCmFinished()
        return render('/casemanagement/digest_body.mako')

    #
    # Delete
    #

    @checkRole('cm')
    def markForDelete(self, id, confirmed):
        id = self._checkInt(id)
        confirmed = self._checkBool(confirmed)
        if confirmed == 1:
            case = self._loadCase(id)
            case.setState(3)
            c.dialog_title = u'Fallakte gelöscht!'
            c.dialog_text = (u"Die Fallakte wurde zum Löschen an die"
                             u" Administration weitergeleitet.")
            c.url_ok = h.url_for(controller="/case_overview")
            session['render_mode'] = 'ro'
            session.save()
            return render('/casemanagement/dialogs/success_delete_cm.mako')
        else:
            c.context = "../main.mako"
            c.dialog_title = DELETE_CONFIRM
            c.dialog_text = MARKDELETE_CONFIRM_TEXT
            c.url_yes = h.url_for(controller='/case', action='markForDelete',
                                  id=id, confirmed=1)
            c.url_no = h.url_for(controller='/case_overview')
            return render('/casemanagement/dialogs/confirm.mako')

    @checkRole('admin')
    def delete(self, id, confirmed):
        id = self._checkInt(id)
        confirmed = self._checkBool(confirmed)
        if confirmed == 1:
            case = self._loadCase(id)
            case.delete()
            c.dialog_title = u'Fallakte gelöscht!'
            c.dialog_text = u'Bitte klicken Sie "OK", um fortzufahren.'
            c.url_ok = h.url_for(controller="/case_overview")
            return render('/casemanagement/dialogs/success_delete.mako')
        else:
            c.context = "../main.mako"
            c.dialog_title = DELETE_CONFIRM
            c.dialog_text = DELETE_CONFIRM_TEXT
            c.url_yes = h.url_for(controller='/case', action='delete',
                                  id=id, confirmed=1)
            c.url_no = h.url_for(controller='/case_overview')
            return render('/casemanagement/dialogs/confirm_delete.mako')

    #
    # Anonymize
    #
    @checkRole('cm')
    def markForAnonymizeFromOverview(self, id, confirmed):
        url = h.url_for(controller="/case_overview")
        return self._markForAnonymize(id, confirmed, url)

    @checkRole('cm')
    def markForAnonymize(self, id, confirmed):
        url = h.url_for(controller='/case', action='digest', id=id)
        return self._markForAnonymize(id, confirmed, url)

    def _markForAnonymize(self, id, confirmed, url):
        id = self._checkInt(id)
        confirmed = self._checkBool(confirmed)
        case = self._loadCase(id)
        # First check if the case is in a state which allows anonymization
        try:
            case.check_anonymizeability()
        except ConsistenceCheckException, e:
            c.dialog_title = MARKANONYMIZE_FAILED
            c.dialog_text = h.literal(e.value)
            c.url_ok = url
            return render('/casemanagement/dialogs/failed_markanonymize.mako')
        # Ok, show confirmation dialog if the user really wants to anonymize the
        # case.
        if confirmed == 1:
            try:
                c.dialog_title = MARKANONYMIZE_SUCCESS
                c.dialog_text = MARKANONYMIZE_SUCCESS_TEXT
                c.url_ok = h.url_for(controller="/case_overview")
                case.make_anonymizable()
                state = case.getState().setState(4)
                session['render_mode'] = 'ro'
                session.save()
                return render('/casemanagement/dialogs/success_markanonymize.mako')
            except SetDataException, e:
                errors = []
                for k, values in e.errors.iteritems():
                    for v in values:
                        errors.append(v.msg)
                c.dialog_title = MARKANONYMIZE_FAILED
                c.dialog_text = MARKANONYMIZE_FAILED_TEXT % "<br>".join(errors) 
                c.url_ok = h.url_for(controller='/case', action='select',
                                     id=case.id, confirmed=1)
                return render('/casemanagement/dialogs/failed_markanonymize.mako')
        else:
            # Get description of the phase which will be set after anonymisation
            phase_name = []
            pdescriptions = g.mpuls_config.get('phases', 'description')[0]
            phase = pdescriptions.get(determineAnonymizedCaseEndPhaseName(case))
            end_phase = determineAnonymizedCaseEndPhase(case)
            for ph, pairs in g.mpuls_config.get('phases', 'pairs')[0].iteritems():
                if str(end_phase) in pairs:
                    phase_name.append(phase)
                    if str(end_phase) == pairs[0]:
                        phase_name.append("(%s)" % _('Running'))
                    else:
                        phase_name.append("(%s)" % _('Finished'))
            
            c.dialog_title = ANONYMIZE_CONFIRM
            c.dialog_text = MARKANONYMIZE_CONFIRM_TEXT % " ".join(phase_name) 
            c.url_yes = h.url_for(controller='/case', action='markForAnonymize',
                                  id=id, confirmed=1)
            c.url_no = url
            return render('/casemanagement/dialogs/confirm_markanonymize.mako')

    @checkRole('admin')
    def anonymize(self, id, confirmed):
        id = self._checkInt(id)
        confirmed = self._checkBool(confirmed)
        case = self._loadCase(id)

        # First check if the case is in a state which allows anonymization
        try:
            case.check_anonymizeability()
        except ConsistenceCheckException, e:
            c.dialog_title = MARKANONYMIZE_FAILED
            c.dialog_text = h.literal(e.value)
            c.url_ok = h.url_for(controller='/case', action='digest', id=id)
            return render('/casemanagement/dialogs/failed_markanonymize.mako')

        # Ok, show confirmation dialog if the user really wants to anonymize the
        # case.
        if confirmed == 1:
            case.anonymize()
            c.dialog_title = ANONYMIZE_SUCCESS
            c.dialog_text = ANONYMIZE_SUCCESS_TEXT
            c.url_ok = h.url_for(controller="/case_overview")
            return render('/casemanagement/dialogs/success_anonymize.mako')
        else:
            c.context = "../main.mako"
            c.dialog_title = ANONYMIZE_CONFIRM
            c.dialog_text = ANONYMIZE_CONFIRM_TEXT
            c.url_yes = h.url_for(controller='/case', action='anonymize',
                                  id=id, confirmed=1)
            c.url_no = h.url_for(controller='/case', action='digest', id=id)
            return render('/casemanagement/dialogs/confirm_anonymize.mako')

    #
    # Restore
    #

    @checkRole('admin')
    def restore(self, id, confirmed):
        id = self._checkInt(id)
        confirmed = self._checkBool(confirmed)
        case = self._loadCase(id)
        if confirmed == 1:
            case.restore()
            c.dialog_title = u'Fallakte wiederhergestellt!'
            c.dialog_text = u'Bitte klicken Sie "OK", um fortzufahren.'
            c.url_ok = h.url_for(controller="/case_overview")
            return render('/casemanagement/dialogs/success_restore.mako')
        else:
            c.context = "../main.mako"
            c.dialog_title = RESTORE_CONFIRM
            c.dialog_text = RESTORE_CONFIRM_TEXT
            c.url_yes = h.url_for(controller='/case', action='restore', id=id,
                                  confirmed=1)
            c.url_no = h.url_for(controller='/case_overview')
            return render('/casemanagement/dialogs/confirm_restore.mako')

    #
    # Printing
    #
    @checkRole(('admin', 'cm', 'pt_dlr'))
    def printdigest(self, id):
        c.print_version = 1
        return self.digest(id)

    @checkRole(('admin', 'cm', 'pt_dlr'))
    def printAll(self, id, empty='0'):
        c.print_version = 1
        out = []
        case = self._loadCase(id)
        c.case = case
        render_mode = session['render_mode']
        session['render_mode'] = 'ro'
        session.save()
        digest = render('/casemanagement/digest_body.mako')
        out.append(digest)
        printed_headings = set()
        navigation = session.get('navigation.tree')
        if navigation:
            for item in navigation.getAllItems():
                # create the headings for all hierarchy elements that
                # have no page of their own.  We can do that by looking
                # at all the parents of the current item that have not
                # yet been used.
                headings = []
                parent = item.parent
                while parent is not None:
                    if parent not in printed_headings:
                        if parent.description:
                            # Ideally, we should be using different
                            # levels of hierarchies, but currently all
                            # headings are H1 headings
                            headings.append("<h1>%s</h1>\n"
                                            % parent.description)
                        printed_headings.add(parent)
                    parent = parent.parent
                out.extend(reversed(headings))

                out.append(get_rendered_page(item.key))
        c.formular = "".join(out)
        session['render_mode'] = render_mode
        session.save()
        return render('/formularpage/formular.mako')

    #
    # Create new case
    #

    def _createCase(self, form_result):
        case = g.case_factory.createNew(data=convert2dic({}))
        case._load_formed_instance_tree()
        # Set data
        data = {}
        for k in convert2dic(form_result).keys():
            try:
                item = case.formed_instance.getAllItemsForName(k)[0]
                id = ":".join(item.getIdentifier().split(':')[1:])
                data["%s:%s" % (k,id)] = form_result.get(k) 
            except:
                # some values within the formresult can not be found in
                # instance tree. e.g "statement_signed"
                pass
            case.formed_instance.setData(data)
            case.formed_instance.commit()

        statement = case.getPrivacyStatement()
        # 3. Do we want to sign the statement?
        if form_result.get('statement_signed'):
            statement.sign()
        return case

    @checkRole('cm')
    def new(self):
        session_case = SessionCase()
        session['case'] = session_case
        session.save()
        redirect_to('new2')

    @checkRole('cm')
    def new2(self):
        c.case = session.get('case')
        return render('/casemanagement/new.mako')

    @checkRole('cm_ka')
    def newAction(self):
        # TODO: Use htmlfill here!
        session_case = SessionCase()
        session_case.setData(request.params)

        session['case'] = session_case
        session.save()

        # Do we want to print the statement?
        if request.params.has_key('print'):
            c.print_version = 1
            statement = PrivacyStatement(session_case, Agency())
            c.content = statement.fillout()
            return render('privacy/default_statement.mako')

        # Or do we want to store the new case?
        elif request.params.has_key('save'):
            validator = NewCaseForm()
            try:
                form_result = validator.to_python(request.params)
                # Create a new case
                case = self._createCase(form_result)
            except formencode.Invalid, error:
                log.debug("Invalid form data: %r", error)
                c.form_result = error.value
                c.form_errors = error.error_dict or {}
                return render('/casemanagement/new.mako')

        return self.select(str(case.id))

    @checkRole('cm')
    def copy(self, id, confirmed):
        id = self._checkInt(id)
        confirmed = self._checkBool(confirmed)
        case = self._loadCase(id)
        if confirmed: 
            case.copy()
            c.dialog_title = _('Case copied!')
            c.dialog_text = _('The case was succussfully copied and is now'
                              ' available in the case overview.'
                              ' Click on "OK" to open the case overview.')
            c.url_ok = h.url_for(controller="/case_overview")
            return render('/casemanagement/dialogs/success_copy.mako')
        else:
            c.dialog_title = _('Copy case?')
            c.dialog_text =  _('Do you really want to copy this case?'
                               ' All data will be copied in the new case.')
            c.url_yes = h.url_for(controller='/case', action='copy',
                                  id=id, confirmed=1)
            c.url_no = h.url_for(controller='/case', action='digest', id=id)
            return render('/casemanagement/dialogs/confirm_copy.mako')

    @checkRole("cm")
    def importCase(self):
        return render('/casemanagement/upload.mako')

    @checkRole('cm_ka')
    def importXMLAction(self):
        try:
            xmlfile = request.POST['xmlfile']
            try:
                importer = MPulsXMLImporter(g.formedTree)
                importer.import_xml(PostgresDBInterface(), xmlfile.file)
            finally:
                xmlfile.file.close()

            c.dialog_title = IMPORT_XML_SUCCESS
            c.dialog_text = IMPORT_XML_SUCCESS_TEXT
            c.url_ok = h.url_for(controller="/case_overview")
            return render('/casemanagement/dialogs/success_xml_import.mako')
        except Exception, e:
            log.exception("Exception during xml import")
            c.dialog_title  = IMPORT_FAILED
            if isinstance(e, XMLImportException):
                c.dialog_text = h.escape(e.message)
            else:
                c.dialog_text = IMPORT_FAILED_TEXT
            c.url_ok = h.url_for(controller='/case', action='importCase')
            return render('/casemanagement/dialogs/failed_xml_import.mako')

    #
    # ORGANISATION (standin and editor settings)
    #

    @checkRole(('admin', 'cm'))
    def organisation(self, id):
        id = self._checkInt(id)
        ulist = UserListObject()
        list = UserGroupList()
        c.case = self._loadCase(id)
        c.standin = c.case.getStandin()
        c.editor = c.case.getEditor()
        c.current_user = session['USER_AUTHORIZED']

        # group list for standin
        filter = [u.gid for u in ulist.getAdminList()]
        filter.append(c.editor.gid)
        filter2 = filter[:]
        filter.append(c.current_user.gid)
        c.group_list = list.getGroupList(filter)
        c.group_list2 = list.getGroupList(filter2)

        # userlist for editor
        filter3 = [u.id for u in ulist.getAdminList()]
        filter3.append(c.current_user.id)
        filter3.append(c.editor.id)
        c.user_list = ulist.getUserList(filter3)
        return render('/casemanagement/organisation.mako')

    @checkRole('admin')
    def setEditorAction(self):
        validator = SetEditorForm()
        try:
            form_result = validator.to_python(request.params)
            case = self._loadCase(form_result.get('case_id'))
            case.setEditor(form_result.get('editor'))
            c.status_msg = h.status_message("Bearbeiter/in erfolgreich gesetzt.")
            return self.organisation(form_result.get('case_id'))
        except formencode.Invalid, error:
            c.form_result = error.value
            c.form_errors = error.error_dict or {}
            c.status_msg = h.status_message("Fehler! Bearbeiter/in nicht gesetzt.")
            return render('/casemanagement/organisation.mako')

    @checkRole(('admin', 'cm'))
    def setStandinAction(self):
        validator = SetStandinForm()
        data = formencode.variabledecode.variable_decode(request.params)
        try:
            form_result = validator.to_python(data)
            case = self._loadCase(form_result.get('case_id'))
            standin = case.getStandin()
            standin.setGroups(form_result.get('standin'))
            c.status_msg = h.status_message("Vertretung erfolgreich gesetzt.")
            return self.organisation(form_result.get('case_id'))
        except formencode.Invalid, error:
            c.form_result = error.value
            c.form_errors = error.error_dict or {}
            c.status_msg = h.status_message("Fehler! Vertretung nicht gesetzt.")
            return render('/casemanagement/organisation.mako')

    @checkRole('cm')
    @checkPrivacyStatement('keepActive')
    def keepActive(self, id, confirmed):
        # If logbook is enabled let the user give the reason why he keeps the
        # case active.
        if g.mpuls_config.is_enabled('case-module', 'logbook'):
            return self._keep_active_with_logbook(id)
        # Just refresh last access date.
        else:
            return self._keep_active_without_logbook(id)

    def _keep_active_with_logbook(self, id):
        entry = LogbookEntry()
        entry.setDate(datetime.now())
        entry.setKind(-2)
        form_defaults = {}
        form_defaults['case_id'] = id 
        form_defaults['date'] = entry.getDate()
        form_defaults['time'] = entry.getTime()
        form_defaults['kind'] = entry.getKindAsInt()
        form_defaults['short_notice'] = u"Aufbewahrung verlängert"
        form_defaults['notice'] = (u"Bitte geben Sie hier den Grund für"
                                   u" die Verlängerung der Aufbewahrungsfrist"
                                   u" ein.")
        id = self._checkInt(id)
        case = self.init_case(id)
        c.entry = entry
        c.form_errors = {}
        form = render('/logbook/new.mako')
        return formencode.htmlfill.render(form, defaults=form_defaults,
                                          auto_insert_errors=False,
                                          errors=c.form_errors)

    def _keep_active_without_logbook(self, id):
        id = self._checkInt(id)
        case = self._loadCase(id)
        try:
            # Temporarily switch the state of the case to force refreshing the last
            # access date.
            state = case.getState().getState()
            case.getState().setState(2)
            case.getState().setState(state)
            c.dialog_title = _('Case storage extended')
            c.dialog_text = _('Please click on "OK" to continue') 
            c.url_ok = h.url_for(controller="/")
            return render('/casemanagement/dialogs/success_keepactive.mako')
        except:
            log.exception('Can not refresh case in "_keep_active_without_logbook"')
            raise

        pass

# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf8:
